from transitions import Machine

import numpy as np
import RobotDART as rd
import dartpy  # OSX breaks if this is imported before RobotDART
import copy
from utils import damped_pseudoinverse, AdT, enforce_joint_limits

class PITask:
    def __init__(self, target, dt, Kp = 10., Ki = 0.1):
        self._target = target
        self._dt = dt
        self._Kp = Kp
        self._Ki = Ki
        self._sum_error = 0

    def set_target(self, target):
        self._target = target

    # function to compute error
    def error(self, tf):
        # 2 ways of computing rotation error in world frame
        # # 1st way: compute error in body frame
        # error_in_body_frame = rd.math.logMap(tf.rotation().T @ self._target.rotation())
        # # transform it in world frame
        # error_in_world_frame = error_in_body_frame @ tf.rotation().T
        # 2nd way: compute error directly in world frame
        rot_error = rd.math.logMap(self._target.rotation() @ tf.rotation().T)
        lin_error = self._target.translation() - tf.translation()
        return np.r_[rot_error, lin_error]

    def update(self, current):
        error_in_world_frame = self.error(current)

        self._sum_error = self._sum_error + error_in_world_frame * self._dt

        return self._Kp * error_in_world_frame + self._Ki * self._sum_error

class MyRobot:
    def __init__(self, dt):
        # Load a robot
        self.robot = rd.Iiwa()
        self.robot.set_actuator_types("servo")
        self.robot.set_position_enforced(False)

        # set initial joint positions
        self.target_positions = copy.copy(self.robot.positions())
        self.target_positions[0] = -2.
        self.target_positions[3] = -np.pi / 2.0
        self.target_positions[5] = np.pi / 2.0
        self.robot.set_positions(self.target_positions)

        # get end-effector pose
        self.eef_link_name = "iiwa_link_ee"
        self.tf_desired = self.robot.body_pose(self.eef_link_name)
        self.vec_desired = self.robot.body_pose_vec(self.eef_link_name)

        self.dt = dt
        self.Kp = 2. # Kp could be an array of 6 values
        self.Ki = 0.01 # Ki could be an array of 6 values
        self.controller = PITask(self.tf_desired, self.dt, self.Kp, self.Ki)

        self.reset()

    def reset(self):
        # set robot to random positions
        self.robot.reset()
        positions = copy.copy(self.target_positions) + np.random.rand(self.robot.num_dofs()) * np.pi / 1.5 - np.pi / 3.
        self.robot.set_positions(enforce_joint_limits(self.robot, positions))

    def update(self):
        tf = self.robot.body_pose(self.eef_link_name)
        vel = self.controller.update(tf)
        jac = self.robot.jacobian(self.eef_link_name) # this is in world frame
        jac_pinv = damped_pseudoinverse(jac) # np.linalg.pinv(jac) # get pseudo-inverse
        cmd = jac_pinv @ vel

        # alpha = 10.
        # cmd = alpha * (jac.T @ vel) # using jacobian transpose
        self.robot.set_commands(cmd)

    def target(self):
        return self.controller._target

    def reached_target(self):
        tf = self.robot.body_pose(self.eef_link_name)
        err = self.controller.error(tf)
        if np.linalg.norm(err) < 1e-3:
            return True
        return False

    def on_enter_A(self):
        print("Entering A")
        self.controller = PITask(self.tf_desired, self.dt, self.Kp, self.Ki)

    def on_enter_B(self):
        print("Entering B")
        tf = dartpy.math.Isometry3()
        tf.set_translation([self.tf_desired.translation()[0], self.tf_desired.translation()[1], self.tf_desired.translation()[2] - 0.2])
        tf.set_rotation(self.tf_desired.rotation())
        self.controller = PITask(tf, self.dt, self.Kp, self.Ki)

    def on_enter_C(self):
        print("Entering C")
        tf = self.tf_desired
        tf = dartpy.math.Isometry3()
        tf.set_translation([self.tf_desired.translation()[0] + 0.2, self.tf_desired.translation()[1], self.tf_desired.translation()[2] - 0.2])
        tf.set_rotation(self.tf_desired.rotation())
        self.controller = PITask(tf, self.dt, self.Kp, self.Ki)

# Create simulator object
dt = 0.001
simu = rd.RobotDARTSimu(dt)

# Create MyRobot
robot = MyRobot(dt)

goal_robot = rd.Robot.create_ellipsoid(dims=[0.15, 0.15, 0.15], pose=robot.vec_desired, color=[0., 1., 0., 0.5], ellipsoid_name="target")

# Create Graphics
gconfig = rd.gui.GraphicsConfiguration(1024, 768) # Create a window of 1024x768 resolution/size
graphics = rd.gui.Graphics(gconfig) # create graphics object with configuration
simu.set_graphics(graphics)
graphics.look_at([3., 1., 2.], [0., 0., 0.])
# graphics.record_video("fsm.mp4")

# Add robot and nice floor
simu.add_robot(robot.robot)
simu.add_visual_robot(goal_robot)
simu.add_checkerboard_floor()

# FSM Machine
# The states
FSM_states = ['A', 'B', 'C']

# And some transitions between states. We're lazy, so we'll leave out
# the inverse phase transitions (freezing, condensation, etc.).
FSM_transitions = [
    { 'trigger': 'arrived', 'source': 'A', 'dest': 'B', 'conditions' : 'reached_target'},
    { 'trigger': 'arrived', 'source': 'B', 'dest': 'C', 'conditions' : 'reached_target'},
    { 'trigger': 'arrived', 'source': 'C', 'dest': 'A', 'conditions' : 'reached_target'}
]

machine = Machine(model=robot, states=FSM_states, transitions=FSM_transitions, initial='A')

# Run simulation
while True:
    if simu.step_world():
        break
    # update visual target object
    goal_robot.set_base_pose(robot.target())
    # update our robot
    robot.update()
    # tick FSM
    robot.arrived()
