from transitions import Machine
import py_trees

import numpy as np
import RobotDART as rd
import dartpy  # OSX breaks if this is imported before RobotDART
import copy
from utils import damped_pseudoinverse, AdT, enforce_joint_limits

class PITask:
    def __init__(self, target, dt, Kp = 10., Ki = 0.1):
        self._target = target
        self._dt = dt
        self._Kp = Kp
        self._Ki = Ki
        self._sum_error = 0

    def set_target(self, target):
        self._target = target

    # function to compute error
    def error(self, tf):
        # 2 ways of computing rotation error in world frame
        # # 1st way: compute error in body frame
        # error_in_body_frame = rd.math.logMap(tf.rotation().T @ self._target.rotation())
        # # transform it in world frame
        # error_in_world_frame = error_in_body_frame @ tf.rotation().T
        # 2nd way: compute error directly in world frame
        rot_error = rd.math.logMap(self._target.rotation() @ tf.rotation().T)
        lin_error = self._target.translation() - tf.translation()
        return np.r_[rot_error, lin_error]

    def update(self, current):
        error_in_world_frame = self.error(current)

        self._sum_error = self._sum_error + error_in_world_frame * self._dt

        return self._Kp * error_in_world_frame + self._Ki * self._sum_error

class ReachTarget(py_trees.behaviour.Behaviour):
    def __init__(self, robot, tf_desired, dt, goal_robot, name="ReachTarget"):
        super(ReachTarget, self).__init__(name)
        # robot
        self.robot = robot
        # end-effector name
        self.eef_link_name = "iiwa_link_ee"
        # set target tf
        self.tf_desired = dartpy.math.Isometry3()
        self.tf_desired.set_translation(tf_desired.translation())
        self.tf_desired.set_rotation(tf_desired.rotation())
        # dt
        self.dt = dt

        self.logger.debug("%s.__init__()" % (self.__class__.__name__))

    def setup(self):
        self.logger.debug("%s.setup()->does nothing" % (self.__class__.__name__))

    def initialise(self):
        self.logger.debug("%s.initialise()->init controller" % (self.__class__.__name__))
        self.Kp = 2. # Kp could be an array of 6 values
        self.Ki = 0.01 # Ki could be an array of 6 values
        self.controller = PITask(self.tf_desired, self.dt, self.Kp, self.Ki)

        goal_robot.set_base_pose(self.controller._target)

    def update(self):
        new_status = py_trees.common.Status.RUNNING
        # control the robot
        tf = self.robot.body_pose(self.eef_link_name)
        vel = self.controller.update(tf)
        jac = self.robot.jacobian(self.eef_link_name) # this is in world frame
        jac_pinv = damped_pseudoinverse(jac) # np.linalg.pinv(jac) # get pseudo-inverse
        cmd = jac_pinv @ vel
        # alpha = 10.
        # cmd = alpha * (jac.T @ vel) # using jacobian transpose
        self.robot.set_commands(cmd)

        # if error too small, report success
        err = np.linalg.norm(self.controller.error(tf))
        if err < 1e-3:
            new_status = py_trees.common.Status.SUCCESS

        if new_status == py_trees.common.Status.SUCCESS:
            self.feedback_message = "Reached target"
            self.logger.debug("%s.update()[%s->%s][%s]" % (self.__class__.__name__, self.status, new_status, self.feedback_message))
        else:
            self.feedback_message = "Error: {0}".format(err)
            self.logger.debug("%s.update()[%s][%s]" % (self.__class__.__name__, self.status, self.feedback_message))
        return new_status

    def terminate(self, new_status):
        self.logger.debug("%s.terminate()[%s->%s]" % (self.__class__.__name__, self.status, new_status))

# Create simulator object
dt = 0.001
simu = rd.RobotDARTSimu(dt)

# Load a robot
robot = rd.Iiwa()
robot.set_actuator_types("servo")
robot.set_position_enforced(False)

# set initial joint positions
target_positions = copy.copy(robot.positions())
target_positions[0] = -2.
target_positions[3] = -np.pi / 2.0
target_positions[5] = np.pi / 2.0
robot.set_positions(target_positions)

# get end-effector pose
eef_link_name = "iiwa_link_ee"
tf_desired = robot.body_pose(eef_link_name)
vec_desired = robot.body_pose_vec(eef_link_name)

goal_robot = rd.Robot.create_ellipsoid(dims=[0.15, 0.15, 0.15], pose=vec_desired, color=[0., 1., 0., 0.5], ellipsoid_name="target")

# Set robot to random initial position
positions = copy.copy(target_positions) + np.random.rand(robot.num_dofs()) * np.pi / 1.5 - np.pi / 3.
robot.set_positions(enforce_joint_limits(robot, positions))

# Create Graphics
gconfig = rd.gui.GraphicsConfiguration(1024, 768) # Create a window of 1024x768 resolution/size
graphics = rd.gui.Graphics(gconfig) # create graphics object with configuration
simu.set_graphics(graphics)
graphics.look_at([3., 1., 2.], [0., 0., 0.])
# graphics.record_video("bt.mp4")

# Add robot and nice floor
simu.add_robot(robot)
simu.add_visual_robot(goal_robot)
simu.add_checkerboard_floor()

# Behavior Tree
py_trees.logging.level = py_trees.logging.Level.DEBUG

# Create tree root
root = py_trees.composites.Parallel(name="Root", policy=py_trees.common.ParallelPolicy.SuccessOnOne())
# Create sequence node (for sequential targets)
sequence = py_trees.composites.Sequence(name="Sequence")
# Target A
trA = ReachTarget(robot, tf_desired, dt, goal_robot, "Reach Target A")
# add target to sequence node
sequence.add_child(trA)
# Target B
tf = dartpy.math.Isometry3()
tf.set_translation([tf_desired.translation()[0], tf_desired.translation()[1], tf_desired.translation()[2] - 0.2])
tf.set_rotation(tf_desired.rotation())
trB = ReachTarget(robot, tf, dt, goal_robot, "Reach Target B")
# add target to sequence node
sequence.add_child(trB)
# Target C
tf = dartpy.math.Isometry3()
tf.set_translation([tf_desired.translation()[0] + 0.2, tf_desired.translation()[1], tf_desired.translation()[2] - 0.2])
tf.set_rotation(tf_desired.rotation())
trC = ReachTarget(robot, tf, dt, goal_robot, "Reach Target C")
# add target to sequence node
sequence.add_child(trC)
# Add sequence to tree
root.add_child(sequence)

# Render tree structure
# py_trees.display.render_dot_tree(root)

# tick once
root.tick_once()

# Run simulation
while True:
    if simu.step_world():
        break
    # update our Behavior Tree to control the robot
    root.tick_once()
